#pragma once
#include <cmath>
#pragma warning(disable: 4068)
#include <chipmunk.h>
using namespace std;

using namespace System;
using namespace System::Drawing;
using namespace System::Drawing::Drawing2D;
using namespace System::Windows::Forms;
using namespace System::Collections;
using namespace System::Collections::Generic;

#include "Color.h"
#include "Sprite.h"
#include "DrawRequest.h"
#include "SpriteLand.h"
using namespace lccs;

void postSolve(cpArbiter* arbiter, cpSpace* space, void* data);



namespace lccs {

    ref class PointerBox
    {
    public:
      PointerBox(Sprite* pointer) : pointer_(pointer) {}
      Sprite* pointer_;
      
    };

	/// <summary>
	/// Summary for SpriteWindow
	///
	/// WARNING: If you change the name of this class, you will need to change the
	///          'Resource File Name' property for the managed resource compiler tool
	///          associated with all .resx files this class depends on.  Otherwise,
	///          the designers will not be able to interact properly with localized
	///          resources associated with this form.
    /// </summary>
	public ref class SpriteWindow : public System::Windows::Forms::Form
	{
	public:
        static SpriteWindow^ getSpriteWindow()
        {
          if(!SpriteWindow::windowExists_)
          {
            theSpriteWindow_ = gcnew SpriteWindow();
            SpriteWindow::windowExists_ = true;
          }
          return theSpriteWindow_;

        }

        void reportCollision(cpShape* shapeA, cpShape* shapeB)
        {
          if (shapeToSprite_->ContainsKey((unsigned int) shapeA) && 
            shapeToSprite_->ContainsKey((unsigned int) shapeB))
          {
            Sprite* firstSprite = (Sprite*) shapeToSprite_[(unsigned int) shapeA];
            Sprite* secondSprite = (Sprite*) shapeToSprite_[(unsigned int) shapeB];

            //firstSprite->onCollision(secondSprite);
            //secondSprite->onCollision(firstSprite);

            firstSprite->addCollision(secondSprite);
            secondSprite->addCollision(firstSprite);
          }
          else
          {
            if (shapeToSprite_->ContainsKey( (unsigned int) shapeA))
            {
              Sprite* firstSprite = (Sprite*) shapeToSprite_[(unsigned int) shapeA];
              //firstSprite->onCollision();
              firstSprite->addCollision(0);
            }
            else
            {

              if (shapeToSprite_->ContainsKey( (unsigned int) shapeB))
              {
                Sprite* secondSprite = (Sprite*) shapeToSprite_[(unsigned int) shapeB];
                //secondSprite->onCollision();
                secondSprite->addCollision(0);
              }
            }
          }
        }

    private:
		SpriteWindow(void)
		{
			InitializeComponent();
			//
			//TODO: Add the constructor code here
			//
            shapeToSprite_ = gcnew Dictionary<unsigned int, unsigned int>;
            this->drawRequests_ = gcnew ArrayList();

            space_ = cpSpaceNew();
            cpSpaceSetGravity(space_, cpv(0 ,0));

            
            const int DEFAULT_WIDTH = 800;
            const int DEFAULT_HEIGHT = 800;

			setWindowSize(DEFAULT_WIDTH, DEFAULT_HEIGHT);
    
            //top_ = cpSegmentShapeNew(space_->staticBody, cpv(0, 0), cpv(DEFAULT_WIDTH, 0), 0);
            //cpShapeSetFriction(top_, .99);
            //cpSpaceAddShape(space_, top_);

            //bottom_ = cpSegmentShapeNew(space_->staticBody, cpv(0, DEFAULT_HEIGHT), cpv(DEFAULT_WIDTH, DEFAULT_HEIGHT), 0);
            //cpShapeSetFriction(bottom_, .99);
            //cpSpaceAddShape(space_, bottom_);

            //left_ = cpSegmentShapeNew(space_->staticBody, cpv(0, 0), cpv(0, DEFAULT_HEIGHT), 0);
            //cpShapeSetFriction(left_, .99);
            //cpSpaceAddShape(space_, left_);

            //right_ = cpSegmentShapeNew(space_->staticBody, cpv(DEFAULT_WIDTH, 0), cpv(DEFAULT_WIDTH, DEFAULT_HEIGHT), 0);
            //cpShapeSetFriction(right_, .99);
            //cpSpaceAddShape(space_, right_);




            const float DEFAULT_REFRESHES_PER_SECOND = 30.0f;
            refreshesPerSecond_ = DEFAULT_REFRESHES_PER_SECOND;


            cpSpaceSetDefaultCollisionHandler( space_, 0, 0, postSolve, 0, 0 );
		}
    public:

        Sprite* getSpriteFromShape(cpShape* shape)
        {
          unsigned int ptrValue = shapeToSprite_[(unsigned int) shape];
          return (Sprite*) ptrValue;
        }

        void setWindowSize(int width, int height)
        {
          cpShapeFree(top_);
          top_ = cpSegmentShapeNew(space_->staticBody, cpv(0, 0), cpv(width, 0), 0);
          cpShapeSetFriction(top_, 0);
          cpSpaceAddShape(space_, top_);
          cpShapeSetElasticity(top_, .99);

          cpShapeFree(bottom_);
          bottom_ = cpSegmentShapeNew(space_->staticBody, cpv(0, height), cpv(width, height), 0);
          cpShapeSetFriction(bottom_, 0);
          cpSpaceAddShape(space_, bottom_);
          cpShapeSetElasticity(bottom_, .99);

          cpShapeFree(right_);
          right_ = cpSegmentShapeNew(space_->staticBody, cpv(width, 0), cpv(width, height), 0);
          cpShapeSetFriction(right_, 0);
          cpSpaceAddShape(space_, right_);
          cpShapeSetElasticity(right_, .99);

          cpShapeFree(left_);
          left_ = cpSegmentShapeNew(space_->staticBody, cpv(0, 0), cpv(0, height), 0);
          cpShapeSetFriction(left_, 0);
          cpSpaceAddShape(space_, left_);
          cpShapeSetElasticity(left_, .99);

          
          int widthDecoration = Width - ClientSize.Width;
          int heightDecoration = Height - ClientSize.Height;

          Width = width + widthDecoration;
          Height = height + heightDecoration;
        }

		cpShape* addSprite(Sprite& newSprite)
		{
			Sprite* spritePtr = newSprite.clone();
            spritePtr->setSpace(space_);
			spriteList_.Add(gcnew PointerBox(spritePtr));
            cpBody* body = spritePtr->getBody();
            cpSpaceAddBody(space_, body);
            cpShape* ballShape = cpSpaceAddShape(space_, cpCircleShapeNew(body, spritePtr->getRadius(), cpvzero));
            spritePtr->addShape(ballShape);
            shapeToSprite_->Add((unsigned int) ballShape, (unsigned int) spritePtr);
            cpShapeSetFriction(ballShape, newSprite.getFriction());
            cpShapeSetElasticity(ballShape, newSprite.getElasticity());

            return ballShape;
		}

        void requestDrawing(DrawRequest^ request)
        {
          this->drawRequests_->Add(request);
        }

		void rotate(float&x, float&y,
			        float centerX, float centerY,
					float theta)
		{
			float xAtCenter = x - centerX;
			float yAtCenter = y - centerY;
			x = xAtCenter*cos(theta) - yAtCenter*sin(theta) + centerX;
			y = yAtCenter*cos(theta) + xAtCenter*sin(theta) + centerY;
		}

        static float distance(float x1, float y1, float x2, float y2)
        {
            return sqrt( (x1-x2)*(x1-x2) + (y1-y2)*(y1-y2) );
        }

        // Compute the distance between two sprites
        static float distance(const Sprite* a, const Sprite* b)
        {
           // Compute (y1-y2) squared
           float sqrY = (a->getYLocation() - b->getYLocation()) *
                        (a->getYLocation() - b->getYLocation());

           // Compute (x1-x2) squared
           float sqrX = (a->getXLocation() - b->getXLocation()) *
                        (a->getXLocation() - b->getXLocation());

           // Distance is sqrt(sqrY+sqrX)
           return sqrt(sqrY + sqrX);
        }


        static bool isColliding(const Sprite* firstSprite, const Sprite* secondSprite)
        {

          // Find the distance between the centers of the 
          // two sprites.
          float centerToCenter = distance(firstSprite, secondSprite);

          // If the centers are closer than the sum of the two radii, 
          // then there is a collision
          return (centerToCenter < 
                 (firstSprite->getRadius() + (secondSprite->getRadius())));
        }

      

          // This function get invoked when the use presses a keyboard key
          void keydownHandler(Object^ sender, KeyEventArgs^ args)
          {
             
            // For each sprite
			List<PointerBox^>^ copyList = copySpriteList();
			for (int i = 0; i < copyList->Count; i++)
            {
              // Extract the sprite from its box
              Sprite* sprite = copyList[i]->pointer_;

              int keyValue = args->KeyValue;

              if (!args->Shift)
              {
                 const int UPPER_TO_LOWER = 32;
                 if (keyValue >= 'A' && keyValue <= 'Z')
                   keyValue = keyValue = keyValue + UPPER_TO_LOWER;
              }

              // Send the value of the key to the sprite
              sprite->onKeyDown(keyValue);
            }
          }


		  List<PointerBox^>^ copySpriteList()
		  {
			  List<PointerBox^>^ spriteListCopy = gcnew List<PointerBox^>();
			  for (int i = 0; i < spriteList_.Count; i++)
			  {
				  spriteListCopy->Add(safe_cast<PointerBox^>(spriteList_[i]));
			  }
			  return spriteListCopy;
		  }

        void mouseClickHandler(Object^ sender, MouseEventArgs^ args)
        {
                 // For each sprite
			// Copy the spriteList
  
			//spritelandMutex_->WaitOne();
			List<PointerBox^>^ copyList = copySpriteList();
          for(int i=0;i<copyList->Count;i++)
          {
             Sprite* sprite = copyList[i]->pointer_;

             float distanceToMouse = distance(static_cast<float>(args->X), 
               static_cast<float>(args->Y), 
                          sprite->getXLocation(), sprite->getYLocation());              

             if (distanceToMouse < sprite->getRadius())
             {
               this->Text = gcnew String(sprite->getName());
               sprite->onMouseClick();
             }
          
           }

		  //spritelandMutex_->ReleaseMutex();
        }

        int getRefreshesPerSecond() 
        {
          return static_cast<int>(refreshesPerSecond_);
        }

        void setRefreshesPerSecond(int newRate)
        {
          float MILLISECONDS_PER_SECOND = 1000.0f;
          refreshesPerSecond_ = static_cast<float>(newRate);
          refreshTimer_->Stop();
          refreshTimer_->Interval =  static_cast<int>(MILLISECONDS_PER_SECOND / static_cast<float>(refreshesPerSecond_));
          refreshTimer_->Start();
        }

        // This function gets invoked every time the screen gets refreshed
        // For this animation, this defaults to 30 times per second.
        void paintHandler(Object^ sender, PaintEventArgs^ args)
        {
			spritelandMutex_->WaitOne();
          // Get a pointer to the window graphics object
          Graphics^ graphics = args->Graphics;

          // Keep a list of sprites that need to be removed.
          ArrayList spritesToRemove;

          // For each sprite
          //for each (Object^ object in spriteList_)
		  Object^ object;

		  List<PointerBox^>^ copyList = copySpriteList();
		  for (int i = 0; i < copyList->Count; i++)
          {
             // Extract the sprite from its box
             Sprite* sprite = copyList[i]->pointer_;

             if (sprite->isAlive())
             {
             // Call the sprite's onDisplay function - send it 
             // the size of the window

             sprite->processCollisions();

             if (sprite->onDisplay(ClientSize.Width, ClientSize.Height))
             {
               int nRequests = drawRequests_->Count;


               for(int i=0;i<nRequests;i++) 
               {
                 DrawRequest^ request = safe_cast<DrawRequest^>(this->drawRequests_[i]);
                 request->draw(graphics, sprite->getXLocation(), 
					 sprite->getYLocation(),
                     sprite->getRotation());
               }

			   drawRequests_->Clear();
             }
             else
             {
                 // if OnDisplay returned false, the sprite wants to be removed 
                 // from the display
                 // Add it to the list of sprites to remove.
                 sprite->setAlive(false);
                 //spritesToRemove.Add(box);
             }
             }
             else
             {
             spritesToRemove.Add(copyList[i]);
             }
          }

          // Remove all the sprites that have requested removal
          for each (Object^ object in spritesToRemove)
          {             
            PointerBox^ box = safe_cast<PointerBox^>(object);
            Sprite* sprite = box->pointer_;
            vector<cpShape*>& shapes = sprite->getShapes();
            for(int i=0;i<shapes.size();i++)
               shapeToSprite_->Remove((unsigned int) shapes[i]);
            //System::Threading::Monitor::Enter(chipmunkMutex_);
            delete sprite;
            spriteList_.Remove(object);
            //System::Threading::Monitor::Exit(chipmunkMutex_);
          }
			spritelandMutex_->ReleaseMutex();
        }

	protected:
		/// <summary>
		/// Clean up any resources being used.
		/// </summary>
		~SpriteWindow()
		{
			if (components)
			{
				delete components;
			}
	}

        // Refresh the screen at every tick event
        void tickEventHandler(Object^ sender, EventArgs^ args)
        {
            // Check for collisions
            //CollisionDetect();
            // System::Threading::Monitor::Enter(chipmunkMutex_);
			spritelandMutex_->WaitOne();
            cpSpaceStep(space_, 1.0/refreshesPerSecond_);
			spritelandMutex_->ReleaseMutex();
            // System::Threading::Monitor::Exit(chipmunkMutex_);

            // Cause the screen to be redrawn 
            Invalidate();
        }


	private:
		/// <summary>
		/// Required designer variable.
		/// </summary>
		System::ComponentModel::Container ^components;
        //array<Sprite*>^ objectList_;
        //vector<Sprite> spriteList_;
        Timer^ refreshTimer_;
        ArrayList spriteList_;
        static SpriteWindow^ theSpriteWindow_;
        static bool windowExists_ = false;
        ArrayList^ drawRequests_;
        cpSpace* space_;
        float refreshesPerSecond_;
        cpShape* top_;
        cpShape* bottom_;
        cpShape* left_;
        cpShape* right_;
        Dictionary<unsigned int, unsigned int>^ shapeToSprite_;
        Object^ chipmunkMutex_;
		System::Threading::Mutex^ spritelandMutex_ = gcnew System::Threading::Mutex();


#pragma region Windows Form Designer generated code
		/// <summary>
		/// Required method for Designer support - do not modify
		/// the contents of this method with the code editor.
		/// </summary>
		void InitializeComponent(void)
		{
			this->components = gcnew System::ComponentModel::Container();
			this->Size = System::Drawing::Size(800,800);
			this->Text = L"SpriteWindow";
			this->Padding = System::Windows::Forms::Padding(0);
			this->AutoScaleMode = System::Windows::Forms::AutoScaleMode::Font;
            chipmunkMutex_ = gcnew Object();
            // Enable double buffering
            this->SetStyle(
              ControlStyles::AllPaintingInWmPaint |
              ControlStyles::UserPaint |
              ControlStyles::DoubleBuffer,true);


            // Install paint handler
            this->Paint += 
              gcnew PaintEventHandler(this, &SpriteWindow::paintHandler);

            // Install keyboard handler
            this->KeyDown += 
               gcnew KeyEventHandler(this, &SpriteWindow::keydownHandler);


            // Install a mouse click handler
            this->MouseClick += 
              gcnew MouseEventHandler(this, &SpriteWindow::mouseClickHandler);

            // Create and configure refresh timer
            refreshTimer_ = gcnew Timer();
            
            const int MILLISECONDS_PER_SECOND = 1000;
            const int DEFAULT_FRAMES_PER_SECOND = 30;
            refreshesPerSecond_ = static_cast<int>(DEFAULT_FRAMES_PER_SECOND);
            
            refreshTimer_->Tick += 
              gcnew EventHandler(this, &SpriteWindow::tickEventHandler);
            refreshTimer_->Interval = 
              static_cast<int>(MILLISECONDS_PER_SECOND / refreshesPerSecond_);
            refreshTimer_->Start();
		}
#pragma endregion
	};
}

